/* SanyoDenki absolute encoder driver for Speedcube3 servo controller

 Pins
P2	ENCTX	encoder transmit data
P3	ENCTE	encoder transmit enable
P4	ENCRX	encoder receive data
*/

#include <propeller2.h>

const int	pinEncTx	= 2;
const int	pinEncTe	= 3;
const int	pinEncRx	= 4;

const int	polyTx	= 0b110;		// CRC3 polynomial x^3+x+1 reversed
const int	polyRx	= 0b10111000;	// CRC8 polynomial x^8+x^4+x^3+x^2+1 reversed

int32_t	position;
int32_t	timeStamp;
int32_t	errorCnt;
int	cog;

void Start ()
{
  cog= cognew (&startEnc, &position) + 1; // @position into PTRA
}

void Stop ()
{
  if (cog) cogstop (cog - 1);
  cog= 0;
}

int32_t GetPosition ()
{
  return position;
}

int32_t GetTime ()
{
  return timeStamp;
}

int32_t GetCrc ()
{
  return errorCnt;
}

/* protocol specification:

For detailed information see document "S5094873B" from SanyoDenki. Encoder
type is "PA035-017BM001" for 17 bits multi-turn absolute data.

Data is transferred according to RS485 standard on a bi-directional 
differential pair (half duplex). Frames have 16 data bits + 1 start + 1 
stop bit = 18 bits alltogether. Baud rate is 2.5MBd and LSB is transferred 
first.

As the P2 shifts the CRC right instead of left the polynomial has to be
reversed.
*/

__pasm {
ORG 0
startEnc
		drvl    #pinEncTe
		fltl	#pinEncTx		' reset smartpin
		wrpin	mode_stx,#pinEncTx	' synchronous serial transmit
		wxpin	baud_bits,#pinEncTx	' 2.5MBd 16N1
		drvl	#pinEncTx		' enable TX smartpin

		fltl	#pinEncRx		' reset smartpin
		wrpin	mode_srx,#pinEncRx	' synchronous serial receive
		wxpin	baud_bits,#pinEncRx	' 2.5MBd 16N1
		' don't enable RX smartpin, it'd read TX data (loopback)

		setse1	#%001_000000 + pinEncRx ' IN rises
txLoop
		drvh	#pinEncTe
		waitx	bitTime			' output enable delay
		wypin	cmdCDF1,#pinEncTx	' start transmission
		waitx	bitTime			' busy flag is not valid before start bit out
.busy		rqpin	rxData,#pinEncTx wc	' read busy flag -> C
	if_c	jmp	#.busy			' wait for TX complete
		waitx	bitTime
		drvl	#pinEncTe
		drvl	#pinEncRx		' enable RX smartpin
		akpin   #pinEncRx		' force IN low
		pollse1				' purge false event
		mov	crc,#$0
		mov	timeout,#0
		getct	maxTime
		add	maxTime,frameTime
		call	#rxFrame		' receive information field
		getct	time	
		mov	status,rxData
		call	#rxFrame		' receive position LSW
		mov	newPos,rxData
		call	#rxFrame		' receive position MSB + CRC

		and	rxData,#$FF
		shl	rxData,#16
		or	newPos,rxData
		cmp	crc,#0 wz		' Z = CRC valid && no timeout
	if_z	cmp	timeout,#0 wz
	if_z	sub	newPos,oldPos
	if_z	mov	lastVel,newPos		' if CRC valid calculate new velocity
		signx	lastVel,#16		'   discard multiturn count, extend to 32 bits
		add	oldPos,lastVel		' else extrapolate position
	if_nz	add	errCnt,#1		'   and increase error count

		setq	#3
		wrlong  oldPos,ptra		' write pos, time, errCnt
		waitx	cycleTime
		jmp     #txLoop

rxFrame		add	maxTime,frameTime
		setq	maxTime
		waitse1	wc			' wait for smartpin
	if_c	mov	timeout,#1
	if_nc	rdpin	rxData,#pinEncRx
		shr	rxData,#16		' right justify data
		mov	rxRev,rxData
		rev	rxRev			' reverse CRC input
		setq	rxRev
		rep	#1,#4
		crcnib	crc,#polyRx		' 4x4=16 bits
		ret

' 			%AAAA_BBBB_FFF_PPPPPPPPPPPPP_TT_MMMMM_0
mode_stx	long	%0000_0000_000_0000000000000_01_11110_0 
mode_srx	long	%0000_0000_000_0000000000000_00_11111_0
baud_bits	long	$10000*72 + 15 ' 2.5MBd 16N1
'			 CRC  CC    EA FC  SC   
cmdCDF1       	long    %100_00001_000_00_010
cycleTime	long	18000 ' 1ms
bitTime		long	72
frameTime	long	72*18

oldPos		long	0
time		long	0
errCnt		long	0
lastVel		long	0

rxData		res	1 ' was RES
newPos		res	1
status		res	1
crc		res	1
rxRev		res	1
maxTime		res	1
timeout		res	1

FIT
}
